/*******************************************************************************
 *                                                                             *
 *        Program : Exclusive Prefix Sum for vector                            *
 *                                                                             *
 *        Author  : Kim, Dongmin (kimd@top.cis.syr.edu)                        *
 *                                                                             *
 *        This program simulates the prefix sum algorithm for vector.          *
 *                                                                             *
 *        The vector size should be multiple of total processor number.        *
 *                                                                             *
 *******************************************************************************/


#include <cm/cmmd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Max   1024
#define fail  0
#define null  0
#define true  1
#define false 0

#define msg_size      4        /* integer */
#define Rankmax       2
#define Bufmax        1024
#define Nmax          1024     /* Maximum global domain size  */
#define Procmax       32       /* Maximum number of processor on each dim*/


/******************************************************************************
                     Definition of Type
 ******************************************************************************/
 
typedef int Datatype; /* data type for array */

typedef int Gindex;   /* for global index */

typedef int Lindex;   /* for local index */

typedef int Pindex;   /* for processor index */

/*typedef enum {false, true} 
     Boolean;*/

/*******************************************************************************
 *                                                                             *
 *   Global variables                                                          *
 *                                                                             *
 *******************************************************************************/
int    my_pnum;            /* The ID of the Processor */
int    my_id;              /* The ID of the selected dim in the converted ID */
int    ori_nproc;          /* Number of available  processors */
int    tot_nproc;          /* Total number of processors */
int    tot_dproc;          /* dimension of tot_nproc */
int    direction;          /* the direction to communicate */
int    dir_nproc;          /* Totol number of processors in the selected dim */
Gindex rank;               /* rank */

int    v_size; /* Vector Size */
int    digit;  /* binary digit */
int    Vector[Max];    /* original data */
int    Original[Max];  /* result data */
int    Sum[Max];     /* Sum from parent process */

int    my_grid[Rankmax];    /* converted ID */
int    nproc[Rankmax];     /* Total number of processors on each dims */
int    dproc[Rankmax];     /* Total dimension of processors */

int    partner_grid[Rankmax];   /* The processor ID with which the data 
				   is exchanged */
int    right_grid[Rankmax];     /* The processor ID of right child */
int    parent_grid[Rankmax];    /* The processor ID of parent */

int    partner_pnum;
int    right_pnum;
int    parent_pnum;

int    width[Rankmax];     /* Width of partition within a single tile */

int    L[Rankmax];         /* Locall Domain size */


/** For Active Message Passing **/
int    rport = 32;
int    prs_dim_ack;
/********************************/


/*******************************************************************************
 *   Procedure :                                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/

/** For Active Message Passing **/
/*******************************************************************************
 *   Procedure : prs_dim_handler                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
void prs_dim_handler() {

  prs_dim_ack = true;

}
/********************************/



show(id, start, size, vector)
     int id, start, size, *vector;
{
  int pn;
  int i;


  CMMD_fset_io_mode(stdout, CMMD_independent);
  for ( pn = 0 ; pn < tot_nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d =>  ", id);

      for ( i = 0 ; i < rank ; i++ )
	printf("%d ", my_grid[i]);
      printf("\n");
	  
      for ( i = start ; i < start+size ; i++ )
	printf("%d  :  %3d\n", i, vector[i]);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}

/*
show2(id, partner, start, size, check)
     int id, partner, start, size, check;
{
  int pn;
  int i;


  CMMD_fset_io_mode(stdout, CMMD_independent);
  for ( pn = 0 ; pn < tot_nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d   Partner %2d   check %d\n", id, partner, check);
      for ( i = start ; i < start+size ; i++ )
	printf("%d  :  %3d\n", i, in_mesg[i]);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}

show3(id, digit, size)
     int id, digit, size;
{
  int pn ;
  int i, j, k, temp_size;


  CMMD_fset_io_mode(stdout, CMMD_independent);

  temp_size = size;
  for ( pn = 0 ; pn < tot_nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d   \n", id);
      for ( i = 0 ; i < digit ; i++ ) {
	for ( j = 0 ; j < v_size ; j++ ) {
	  printf("Right[%d][%d]  :  %3d\n", i, j, Right[i][j]);
	  temp_size = temp_size / 2;
	}
      }

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}

show4(id, ch, value)
     char* ch;
     int value, id;
{

  int    pn;

  CMMD_fset_io_mode(stdout, CMMD_independent);

  for ( pn = 0 ; pn < tot_nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d   \n", id);
	  printf("%s : %d\n", ch, value);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}
*/

/*******************************************************************************
 *   Procedure : read_input                                                    *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
read_input() {
   int error;
   int i, j;
   int total = 1, t_total=1;
   int save;

   ori_nproc = CMMD_partition_size();

   /*---------------- Rank of the input Array ---------------------------*/
   do {
      printf("\nEnter rank of the input array (Max=%d): ", Rankmax);
      scanf("%d", &rank);
      error = ( rank < 1 || rank > Rankmax );
      if (error)
	 printf("\n??? Improper rank. Try again!\n");
    } while (error);

   printf("\nEnter direction to communicate (0 ~ %d): ", rank-1);
   scanf("%d", &direction);

   printf("\nEnter vector size : ");
   scanf("%d", &v_size);

   /*-------------------- from dim=0 to dim=rank-1 ----------------------*/
   for (i=0; i<rank; ++i) {

     /*------------ Source Distribution for A, M ------------------*/
     save = total;
     do {
       printf("\n   Enter number of source processors (Max=%d): ", 
	      Procmax/total);
       scanf("%d", &nproc[i]);
       total *= nproc[i];
       error = ( nproc[i] < 1 || total > ori_nproc );
/*
       error = ( !Blog2(nproc[i])  || nproc[i] < 1 ||
		total > ori_nproc );
*/
       dproc[i] = Ilog2(nproc[i]);
       if (error) {
	 total = save;
	 printf("\n??? Improper source processor number. Try again!\n");
       }
     } while (error);
   }

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);


   tot_nproc = total;
   tot_dproc = Ilog2(tot_nproc);

   /*======= Who I am ? =======*/
   my_pnum  = CMMD_self_address(); 
   map_to_grid(my_pnum, my_grid);

   /* setup the partner, right child and the parent */
   for ( i = 0 ; i < rank ; i++ ) {
     partner_grid[i] = my_grid[i];
     right_grid[i] = my_grid[i];
     parent_grid[i] = my_grid[i];
   }

/*  show(my_pnum, 0, v_size);*/

   i = CMMD_reset_partition_size(tot_nproc);
   if (i != tot_nproc) 
     CMMD_error("\nERROR: CMMD Partition failed!\n");

   /* every processor use a different random seed */
   /*srandom(CMMD_self_address()); */

   if ( CMAML_allocate_this_rport(rport) <= 0 )
     CMMD_error("ERROR: Failed to allocate receive rport %d\n", rport);

}

/*******************************************************************************
 *   Procedure : add_n_store                                                   *
 *   Input     : int start_index, int data_size, int step, int *v_1, int *v_2  *
 *               int *v_3[]                                                    *
 *   Output    : None                                                          *
 *   Function  : Add the two arrays from the start_index by data_size, and     *
 *                store the right child's data.                                *
 *******************************************************************************/
add_n_store(start_index, data_size, step, v_1, v_2, v_3)
     int start_index,
         data_size,
         step,
         *v_1, *v_2, v_3[Max][Max];
{
  int i;

  if ( right_pnum == partner_pnum ) {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      v_1[i] += v_2[i];
      v_3[step][i] = v_2[i];
    }
  }
  else {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      v_3[step][i] = v_1[i];
      v_1[i] += v_2[i];
    }
  }
}

/*******************************************************************************
 *   Procedure : Ilog2()                                                       *
 *   Input     : int n                                                         *
 *   Output    : log2(n)                                                       *
 *   Function  : return the integer value of log2(n) if it is integer.         *
 *               If not, return -1.                                            *
 *******************************************************************************/
int Ilog2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1 << i != n )
    return -1;
  else
    return i;
}

/*******************************************************************************
 *   Procedure : pow2()                                                        *
 *   Input     : int x                                                         *
 *   Output    : 2^x                                                           *
 *   Function  : power of 2 for the input value                                *
 *******************************************************************************/
int pow2(x)
     int x;
{
  int p = 1;
  int i;

  for ( i = 0 ; i < x ; i++ )
    p <<= 1;

  return p;
}



/*******************************************************************************
 *   Procedure : map_to_pnum                                                   *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
map_to_pnum(grid, pn)    /* grid to 1-d array */
     int grid[Rankmax];      
     int *pn;
{
  int i, mult=1;

  *pn = 0;

  for (i=0; i<rank; ++i) {
    *pn += grid[i] * mult;
    mult *= nproc[i];
  }

}

/*******************************************************************************
 *   Procedure : map_to_grid                                                   *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
map_to_grid(pn, grid)   /* 1-D array to grid */
     int pn;
     int grid[Rankmax];
{
  int i, mult, mod;

  mult  = tot_nproc;
  mod = pn;

  for (i=rank-1; i>=0; --i) {
    mult /= nproc[i];
    grid[i] = mod / mult;
    mod %= mult;
  }

}


/*******************************************************************************
 *   Procedure :                                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
prs_n(vector_size, input_v, p_result, r_result)
     int vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_v,     /* input array */
         *p_result,    /* prefix sum output array */
         *r_result;    /* reduction output array */
{
  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    store_recv; /* the receive point after the sum */
  int    store_size; /* the data_size after the sum */
  int    Right[Max][Max];   /* Sums from right child process by step.
			       1st dimension is for step, and
			       2nd dimension is for the sums
			       from the right child process.          */
  int    in_mesg[Max];      /* received data */

  /** For Active Message Passing **/
  CMAML_disable_interrupts();
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();

  for ( i = 0 ; i < vector_size ; i++ )
    p_result[i] = input_v[i];

  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(dir_nproc);

/* Sum */

  for ( step = 0 ; step < digit ; step++ ) {
    data_size = data_size / 2;
    start_send = start_recv;
    check = my_id & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    right_pnum   = my_id | pow2(step);
    right_grid[direction] = right_pnum;
    map_to_pnum(right_grid, &right_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &in_mesg[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &p_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    add_n_store(start_recv, data_size, step, p_result, in_mesg, Right);
  }

  store_recv = start_recv;
  store_size = data_size;

  /* copy the p_result to r_result for the reduction */
  for ( i = 0 ; i < vector_size ; i++ )
    r_result[i] = p_result[i];
  show(my_pnum, 0, vector_size, p_result);
/* ===============
     PREFIX SUM
   =============== */

/* Distribution */

  start_send = start_recv;
  for ( step = digit - 1 ; step >= 0 ; step-- ) {
    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner and the right child */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    right_pnum   = my_id | pow2(step);
    right_grid[direction] = right_pnum;
    map_to_pnum(right_grid, &right_pnum);

    if ( partner_pnum == right_pnum ) {

      /** For Active Message Passing **/

      prs_dim_ack = false;

      /* receive */
      CMAML_set_rport_addr         (rport, &p_result[start_recv]);
      CMAML_set_rport_byte_counter (rport, data_size*size);
      CMAML_set_rport_total_nbytes (rport, data_size*size);

      CMMD_sync_with_nodes();

      /* send */
      CMAML_scopy(partner_pnum,
		  rport,
		  0,
		  0,
		  &p_result[start_send],
		  data_size*size,
		  null,
		  null);  

      CMAML_poll_while_lt(&prs_dim_ack, true); 

      /********************************/

      for ( i = start_send ; i < start_send + data_size ; i++ )
	p_result[i] -= Right[step][i];

    }
    else {
      for ( i = start_send ; i < start_send + data_size ; i++ )
	in_mesg[i] = p_result[i] - Right[step][i];

      /** For Active Message Passing **/

      prs_dim_ack = false;

      /* receive */
      CMAML_set_rport_addr         (rport, &p_result[start_recv]);
      CMAML_set_rport_byte_counter (rport, data_size*size);
      CMAML_set_rport_total_nbytes (rport, data_size*size);

      CMMD_sync_with_nodes();

      /* send */
      CMAML_scopy(partner_pnum,
		  rport,
		  0,
		  0,
		  &in_mesg[start_send],
		  data_size*size,
		  null,
		  null);  

      CMAML_poll_while_lt(&prs_dim_ack, true); 

      /********************************/
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

  for ( i = 0 ; i < vector_size ; i++ )
    p_result[i] -= Original[i];

  show(my_pnum, 0, vector_size, p_result);

/* ===============
     REDUCTION
   =============== */

/* all-to-all BC */

  data_size  = store_size;
  start_recv = store_recv;
  start_send = start_recv;

  for ( step = digit - 1 ; step >= 0 ; step-- ) {
    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &r_result[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		&r_result[start_send],
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

  show(my_pnum, 0, vector_size, r_result);
}


/*******************************************************************************
 *   Procedure :                                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/


/*******************************************************************************
 *                                                                             * 
 *                              MAIN  PROGRAM                                  *
 *                                                                             *
 *******************************************************************************/
  
main(argc, argv)
     int argc;
     char *argv[];
{
  FILE   *fd; /* input vector file name */

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  double time, time0, time1, time2;
  int p_result[Max], r_result[Max];
  int    input_v[Max];
  double cmmd, prs;


  /*-------------------------  Initialize  -----------------------------*/

  read_input();

  my_id = my_grid[direction]; 

  dir_nproc  = nproc[direction];

/* Initialize the vector in each processor */

  for ( i = 0 ; i < v_size ; i++ ) {
    Original[i] = 2;
    input_v[i]  = Original[i];
  }


  CMMD_node_timer_clear(0);

  CMMD_node_timer_start(0);
  prs_n(v_size, input_v, p_result, r_result);
  CMMD_node_timer_stop(0);

  time0 = CMMD_node_timer_busy(0) * 1000.;  /* milisec */

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n*********************************************************\n");
  printf("\n  Nproc       = %d", tot_nproc);
  printf("\n  Vector size = %d\n", v_size);
  printf("\n  CMMD Vector exclusive total     = %5.3f msec", time0);
  printf("\n\n*********************************************************\n\n");

/*
  CMMD_sync_with_nodes();
  CMMD_node_timer_clear(0);
  CMMD_node_timer_clear(1);

  CMMD_node_timer_start(0);
  PRS_dim0(v_size, p_result, r_result);
  CMMD_node_timer_stop(0);

  CMMD_node_timer_start(1);
  CMMD_PRS0(v_size, p_result, r_result);
  CMMD_node_timer_stop(1);


  prs = CMMD_node_timer_busy(0) * 1000.; */ /* milisec */
/*  cmmd= CMMD_node_timer_busy(1) * 1000.; */ /* milisec */

/*
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n*******************************************\n");
  printf("\n  Nproc       = %d", nproc);
  printf("\n  Vector size = %d\n", v_size);
  printf("\n  CMMD Vector Scan/Reduction = %5.3f msec", cmmd);
  printf("\n  PRS on dim0                = %5.3f msec \n", prs);
  printf("\n*******************************************\n\n");
  CMAML_enable_interrupts();
*/
}

